/**
 * Copyright  2008 Kevin Twidle, Imperial College, London, England.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Contact: Kevin Twidle <kpt@doc.ic.ac.uk>
 *
 * Created on Sep 8, 2009
 *
 * $Log:$
 */

package net.ponder2.managedobject;

import java.util.List;
import java.util.Vector;

import net.ponder2.ManagedObject;
import net.ponder2.apt.Ponder2op;
import net.ponder2.exception.Ponder2ArgumentException;
import net.ponder2.exception.Ponder2Exception;
import net.ponder2.objects.P2Array;
import net.ponder2.objects.P2Block;
import net.ponder2.objects.P2Hash;
import net.ponder2.objects.P2Object;

/**
 * TODO Description
 * 
 * @author Kevin Twidle
 * @version $Id:$
 */
public class TrRule implements ManagedObject {

	// For testing
	private static final boolean Trace = false;
	private static final boolean SingleThread = false;

	final private P2Block condition;
	// A parallel list of lists of TrActions
	final List<List<TrAction>> actions;
	List<TrAction> currentActions;
	final List<ActionThread> threads;
	private boolean running;
	private String threadName = "NoName";

	/**
	 * Creates a rule with the given condition block aCondition
	 * 
	 * @param aCondition
	 */
	@Ponder2op("condition:")
	TrRule(P2Block aCondition) {
		this.condition = aCondition;
		actions = new Vector<List<TrAction>>();
		threads = new Vector<ActionThread>();
		newActionList();
		running = false;
	}

	private void newActionList() {
		currentActions = new Vector<TrAction>();
		actions.add(currentActions);
	}

	/**
	 * Creates a rule with the given condition block aCondition and the given
	 * action block anAction
	 * 
	 * @param aCondition
	 * @param anAction
	 */
	@Ponder2op("condition:action:")
	TrRule(P2Block aCondition, P2Object anAction) {
		this(aCondition);
		currentActions.add(new TrAction(anAction));
	}

	/**
	 * Creates a rule with the given condition block aCondition and the given
	 * array of blocks in anActionArray
	 * 
	 * @param aCondition
	 * @param anAction
	 * @throws Ponder2ArgumentException
	 */
	@Ponder2op("condition:actions:")
	TrRule(P2Block aCondition, P2Array anActionArray)
			throws Ponder2ArgumentException {
		this(aCondition);
		actions(anActionArray);
	}

	/**
	 * Adds an action to the rule
	 * 
	 * @param anAction
	 */
	@Ponder2op("action:")
	protected void addAction(P2Object anAction) {
		currentActions.add(new TrAction(anAction));
	}

	/**
	 * Adds an array of actions to the rule. The actions are copied from the
	 * array, the array is discarded
	 * 
	 * @param anActionArray
	 * @throws Ponder2ArgumentException
	 */
	@Ponder2op("actions:")
	protected void actions(P2Array anActionArray)
			throws Ponder2ArgumentException {
		for (P2Object action : anActionArray.asArray()) {
			currentActions.add(new TrAction(action));
		}
	}

	/**
	 * Indicates that any following actions are to be run in parallel with the
	 * previously given actions. This command may be used many times for
	 * multiple parallelism
	 * 
	 * @throws Ponder2ArgumentException
	 */
	@Ponder2op("parallel")
	protected void parallel() throws Ponder2ArgumentException {
		newActionList();
	}

	/**
	 * checks the condition of the rule and returns whether it can be run. Typically used by the rulseset
	 * 
	 * @param source
	 * @param thePercepts
	 * @return
	 * @throws Ponder2ArgumentException
	 * @throws Ponder2Exception
	 */
	@Ponder2op("canRun:")
	protected boolean canRun(P2Object source, P2Hash thePercepts)
			throws Ponder2ArgumentException, Ponder2Exception {
		return condition.operation(source, "valueVars:", thePercepts)
				.asBoolean();
	}

	/**
	 * runs the rule's actions one by one. May start one or more concurrent threads if the <b>parallel</b> command has been used
	 * 
	 * @param source
	 * @param thePercepts
	 * @throws Ponder2ArgumentException
	 * @throws Ponder2Exception
	 */
	@Ponder2op("valueVars:")
	protected void run(P2Object source, P2Hash thePercepts)
			throws Ponder2Exception {
		Ponder2Exception error = null;
		threadName = Thread.currentThread().getName();
		// trace("Running rule "+this);
		running = true;
		if (!SingleThread && actions.size() == 1) {
			ActionThread thread = new ActionThread(threadName, source, actions
					.get(0), thePercepts);
			threads.add(thread);
			// Don't start it as a thread!
			thread.run();
			return;
		}
		int index = 0;
		for (List<TrAction> actionList : actions) {
			if (!running)
				break;
			ActionThread thread = new ActionThread(threadName + " action "
					+ ++index, source, actionList, thePercepts);
			threads.add(thread);
			thread.start();
		}

		// Wait for all to finish
		while (!threads.isEmpty()) {
			ActionThread thread = threads.get(0);
			try {
				thread.join();
				error = thread.getError();
				threads.remove(thread);
			} catch (InterruptedException e) {
				trace("TrRule action wait interrupted - ignored");
			}
		}
		// If a thread had an error, throw it now after all threads have
		// finished
		if (error != null)
			throw error;
	}
	
	/**
	 * Tells the rule to stop executing. If an action block is in progress it
	 * will be completed. If a ruleset is in progress it will be sent a stop
	 * message
	 * 
	 * @param source
	 * @throws Ponder2Exception
	 */
	@Ponder2op("stop")
	protected void stop(P2Object source) throws Ponder2Exception {
		if (running) {
			running = false;
			trace("TrRule: stop received");
			for (ActionThread thread : threads) {
				thread.stop(source);
			}
		}
	}

	/**
	 * TODO Description
	 * 
	 * @author Kevin Twidle
	 * @version $Id:$
	 */
	private final class ActionThread extends Thread {
		P2Object source;
		List<TrAction> myActions;
		private TrAction currentAction;
		private P2Hash percepts;
		private Ponder2Exception error = null;

		/**
		 * @param source2
		 * @param actions2
		 * @param thePercepts
		 */
		public ActionThread(String threadName, P2Object source,
				List<TrAction> actions, P2Hash thePercepts) {
			super(threadName);
			this.source = source;
			this.myActions = actions;
			this.percepts = thePercepts;
			currentAction = null;
		}

		@Override
		public void run() {
			for (TrAction action : myActions) {
				if (!running)
					break;
				currentAction = action;
				try {
					action.run(source, percepts);
				} catch (Ponder2Exception e) {
					error = e;
				}
			}
		}

		public void stop(P2Object source) throws Ponder2Exception {
			if (currentAction != null)
				currentAction.stop(source);
		}

		/**
		 * @return the error
		 */
		public Ponder2Exception getError() {
			return error;
		}
	}

	// Class to make simple actions and rule sets look the same
	class TrAction {

		private P2Object action;
		private boolean isBlock;

		TrAction(P2Object action) {
			this.action = action;
			isBlock = (action instanceof P2Block);
		}

		void run(P2Object source, P2Hash percepts) throws Ponder2Exception {
			action.operation(source, "valueVars:", percepts);
			// Testing, slow down
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				trace("Action sleep interrupted - ignoring");
			}

		}

		void stop(P2Object source) throws Ponder2Exception {
			if (!isBlock)
				action.operation(source, "stop");
		}
	}

	private void trace(String msg) {
		if (TrRuleSet.Trace || Trace)
			System.out.println("(" + threadName + ") " + msg);
	}
}
