/**
 * Copyright  2008 Kevin Twidle, Imperial College, London, England.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Contact: Kevin Twidle <kpt@doc.ic.ac.uk>
 *
 * Created on Jul 28, 2009
 *
 * $Log:$
 */

package net.ponder2.managedobject;

import java.util.Vector;

import net.ponder2.ManagedObject;
import net.ponder2.apt.Ponder2op;
import net.ponder2.exception.Ponder2ArgumentException;
import net.ponder2.exception.Ponder2Exception;
import net.ponder2.objects.P2Hash;
import net.ponder2.objects.P2Object;

/**
 * A TrRuleSet manages an ordered set of TrRules. When activated, the rules are
 * asked one by one if they can be run. The first one to return true gets run.
 * If a lower priority rule was running then that rule is stopped before the new
 * one is started. If the new rule is already running then it is left alone.
 * 
 * The rules are reevaluated when a rule finishes or when an event is received.
 * 
 * The TrRuleSet may be told to wait for the currently running rule to complete
 * before scheduling another.
 * 
 * @author Kevin Twidle
 * @version $Id:$
 */
public class TrRuleSet implements ManagedObject {

	Vector<P2Object> rules;
	boolean running = false;
	private boolean waitForRule = false;
	private boolean runOnce = false;
	private final String name;
	Thread myThread = null;
	Thread ruleThread = null;
	private P2Object currentRule = null;
	private String threadName = "NoName";
	// Event handling
	private boolean waitForEvent = false;
	private boolean eventReceived = false;

	public static boolean Trace = false;

	/**
	 * Creates a new TrRuleSet
	 * 
	 */
	@Ponder2op("create")
	TrRuleSet() {
		this("RuleSet");
	}

	/**
	 * Creates a new TrRuleSet with the name aName (used for trace messages)
	 * 
	 */
	@Ponder2op("create:")
	TrRuleSet(String aName) {
		rules = new Vector<P2Object>();
		running = false;
		name = aName;
	}

	/**
	 * Adds aRule to the end of the ordered rule set.
	 * 
	 * @param aRule
	 *            the rule to be added
	 */
	@Ponder2op("add:")
	protected void addRule(P2Object aRule) {
		rules.add(aRule);
	}

	/**
	 * Adds aRule at position anIndex to the rule set. Displaced rules get
	 * pushed down by one
	 * 
	 * @param anIndex
	 * @param aRule
	 */
	@Ponder2op("at:put:")
	protected void atput(int anIndex, P2Object aRule) {
		rules.add(anIndex, aRule);
	}

	/**
	 * Tells the rule set to wait for completion of a rule rather than
	 * interrupting it as necessary if aBoolean is true. The initial value is
	 * false
	 * 
	 * @param aBoolean
	 */
	@Ponder2op("waitForRule:")
	protected void waitForRule(boolean aBoolean) {
		this.waitForRule = aBoolean;
	}

	/**
	 * Tells the rule set to wait for an event before running through the rule
	 * list again if aBoolean is true. Use of this option can remove unnecessary
	 * scans of the rule conditions. The initial value is false.
	 * 
	 * @param aBoolean
	 *            true if an event should be awaited before continuing
	 */
	@Ponder2op("waitForEvent:")
	protected void waitForEvent(boolean aBoolean) {
		this.waitForEvent = aBoolean;
	}

	/**
	 * Tells the rule set to run one rule and then terminate if aBoolean is
	 * true. The initial value is false
	 * 
	 * @param aBoolean
	 */
	@Ponder2op("runOnce:")
	protected void runOnce(boolean aBoolean) {
		this.runOnce = aBoolean;
	}

	/**
	 * Tells the rule set to stop executing rules. If a rule is running then
	 * that rule is told to stop. This action returns after the currently
	 * running rule (if any) has been stopped
	 * 
	 * @throws Ponder2Exception
	 */
	@Ponder2op("stop")
	protected void stop(P2Object source) throws Ponder2Exception {
		if (myThread == null)
			return;
		trace(threadName, "TrRuleset: stopping");
		running = false;
		stopRule(source);
		myThread.interrupt();
		trace(threadName, "TrRuleset: stopped");
	}

	/**
	 * Tells the receiver to reevaluate the rule conditions and to choose a
	 * (possibly) new rule to run
	 * 
	 * @param source
	 */
	@Ponder2op("event")
	protected void event(P2Object source) {
		eventReceived = true;
		trace(threadName, "event received");
		if (myThread == null)
			return;
		myThread.interrupt();
	}

	/**
	 * runs the ruleset using the percepts found in thePercepts. The rule set is
	 * run in a separate thread so that it can be given new events or maybe
	 * stopped. This call immediately returns leaving the ruleset running.
	 * 
	 * @param percepts
	 * @throws Ponder2Exception
	 * @throws Ponder2ArgumentException
	 */
	@Ponder2op("run:")
	protected void run(final P2Object source, final P2Hash thePercepts) {
		Thread rulesThread = new Thread("") {

			@Override
			public void run() {
				try {
					valueVars(source, thePercepts);
				} catch (Ponder2ArgumentException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (Ponder2Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		};
		rulesThread.start();
	}

	/**
	 * runs the ruleset using the percepts found in thePercepts. This is a
	 * synchronous call and returns only when this thread has been told to stop.
	 * Normally the <b>run</b> operation above is used to start the ruleset.
	 * <p>
	 * This operation has this name to make it look the same as running a block
	 * with a variable hash
	 * 
	 * @param source
	 * @param thePercepts
	 * @throws Ponder2ArgumentException
	 * @throws Ponder2Exception
	 */
	@Ponder2op("valueVars:")
	protected void valueVars(P2Object source, P2Hash thePercepts)
			throws Ponder2Exception {
		myThread = Thread.currentThread();
		ruleThread = null;
		currentRule = null;
		running = true;
		int currentRuleNumber = 0;
		threadName = Thread.currentThread().getName() + " " + name + " ";
		while (running) {
			int ruleNumber = 1;
			eventReceived = false;
			// trace("Checking rules");
			for (P2Object rule : rules) {
				if (!running)
					break;
				 trace(threadName, "Checking rule " + ruleNumber);
				if (rule.operation(source, "canRun:", thePercepts).asBoolean()) {
					// We don't want to interfere if the first satisfied rule is
					// the current rule
					trace(threadName, "Want to run rule " + ruleNumber);
					if (rule == currentRule) {
						trace(threadName, "Leaving current rule ("
								+ ruleNumber + ") running");
						break;
					}
					if (currentRuleNumber > 0) {
						trace("(" + threadName + ") Stopping rule "
								+ currentRuleNumber);
						stopRule(source);
					}
					 trace(threadName, "Running rule " + ruleNumber);
					currentRule = rule;
					currentRuleNumber = ruleNumber;
					ruleThread = runRule(source, rule, ruleNumber, thePercepts);
					break;
				}
				ruleNumber++;
			}

			if (ruleThread == null) {
				try {
					trace(threadName, "Ruleset: Error, no valid rule found to run - sleeping");
					while (true)
						Thread.sleep(6000000L);
				} catch (InterruptedException e) {
				}
			} else {
				// Wait for the rule to finish
				while (true) {
					try {
						trace("(" + threadName + ") Awaiting rule thread");
						ruleThread.join();
						currentRule = null;
						currentRuleNumber = 0;
						trace(threadName, "Rule thread joined");
						if (runOnce)
							running = false;
						break;
					} catch (InterruptedException e) {
						trace(threadName, "Ruleset woken up");
						// If we can switch rules on the fly break out of the
						// wait loop
						if (!waitForRule)
							break;
					}
				}
			}
			// Event handling, don't proceed if we have to wait for an event
			while (waitForEvent && !eventReceived) {
				trace(threadName, "Awaiting event");
				try {
					Thread.sleep(10000);
				} catch (InterruptedException e) {
				}
			}
		}
		trace(threadName, "Ruleset finishing");
		stopRule(source);
	}

	private void stopRule(P2Object source) throws Ponder2Exception {
		if (currentRule != null) {
			currentRule.operation(source, "stop");
			while (true)
				try {
					ruleThread.join();
					break;
				} catch (InterruptedException e) {
					trace(threadName,
							"Unexpected interrupt waiting for rule to stop");
				}
		}
	}

	private Thread runRule(final P2Object source, final P2Object rule,
			final int number, final P2Hash thePercepts) throws Ponder2Exception {
		final String name = threadName + " Rule " + number;
		Thread thread = new Thread(name) {

			@Override
			public void run() {
				try {
					trace(name, "starting");
					rule.operation(source, "valueVars:", thePercepts);
					trace(name, "finishing");
				} catch (Ponder2Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		};
		thread.start();
		return thread;
	}

	/**
	 * Sets all TR tracing on for all rulesets and rules if aBoolean is true.
	 * Tracing is set to false initially
	 * 
	 * @param aBoolean
	 *            true to turn tracing on
	 */
	@Ponder2op("trace:")
	protected void trace(boolean aBoolean) {
		Trace = aBoolean;
	}

	private void trace(String threadName, String msg) {
		trace("(" + threadName + ") " + msg);
	}

	private void trace(String msg) {
		if (Trace)
			System.out.println(msg);
	}

}
